//*******************************************************
//	FM channel Radio
//	FM tuner module : SPK-FM-301 / SPK-TFM-1010
//
//	MCU:	ATmega8
//			hfuse:C1h	EEPROM消去保護
//			lfuse:A4h	内蔵CR 8MHz BOD 2.7V
//
//	Compiler:	AVR-GCC		4.3.0
//				AVR-LIBC	1.6.2
//				binutils	2.18
//
//	2009.11.20	SW3+SW4で固定周波数モード、スイッチ操作関係のルーチン見直し
//	2009.10.18	TWI_recvの ACKと NACK取り違えバグ修正
//	2009.8.17	KAWAKAMI Yukio
//*******************************************************

#include	<avr/io.h>
#include	<avr/interrupt.h>
#include	<avr/pgmspace.h>
#include	<avr/eeprom.h>
#include	<util/twi.h>
#include	<string.h>

#include	"FK1.h"
#include	"AR1000.h"

#define		CPU_CLOCK	8000000L	// CPUクロック 8MHz
#define		CH_N		4			// チャンネル数

#define		SLEEP_MIN	30	// スリープモードで電源OFFする時間(分)

#define		R3_VOLUME	R3_VOLUMN_6	// AR1000 出力ボリューム

#define		RSSI_MIN_LEVEL	7	// 有効な電波のレベル

#define		LOW_BATT	190		// ローバッテリー電圧 1.90V

#define		FREQ_EXT	896;	// SW3+SW4で受信する周波数 89.6MHz

#ifdef PRR	// ATmega88/168
#define		ADC_REF	110		// A/Dリファレンス電圧 1.1V
#else		// ATmega8
#define		ADC_REF	256		// A/Dリファレンス電圧 2.56V
#endif

#define		R8_VALUE	3300L	// 抵抗 R8の値
#define		R9_VALUE	10000L	// 抵抗 R9の値


ushort	R1Default;

volatile uchar	Timer;
volatile uchar	FlashTimer;
volatile uchar	TwiTimer;
volatile uchar	FlashLED;
volatile uchar	FMch;
volatile uchar	SwFlag;
volatile uchar	LEDdata;
volatile uchar	SleepFlag;
volatile uchar	SleepTimerDS;
volatile uchar	SleepTimerS;
volatile uchar	SleepTimerM;
volatile ushort	BattVolt;


//*******************************************************
//	タイマー２比較Ａ比較一致割り込み
//	1MHz / 64 / 156 = 100Hz
//*******************************************************
#ifndef TIMER2_COMPA_vect
#define	TIMER2_COMPA_vect	TIMER2_COMP_vect
#endif

ISR(TIMER2_COMPA_vect)
{
	uchar	d;

	// 各種タイマ
	if (Timer)	--Timer;
	if (FlashTimer)	--FlashTimer;
	if (TwiTimer)	--TwiTimer;
	if (SleepFlag){
		if (++SleepTimerDS >= 100){
			SleepTimerDS = 0;
			if (++SleepTimerS >= 60){
				SleepTimerS = 0;
				++SleepTimerM;
			}
		}
	}

	// スイッチ入力値を SwFlagにセット
	SwFlag = 0;
	d = ~PIND;

	if (d & (1<<SW0))		SwFlag |= SW_POWER;
	if (d & (1<<SW1))		SwFlag |= SW_1;
	if (d & (1<<SW2))		SwFlag |= SW_2;
	if (d & (1<<SW3))		SwFlag |= SW_3;
	if (d & (1<<SW4))		SwFlag |= SW_4;

	// LED点滅処理
	if (FlashLED){
		if (FlashTimer == 0){
			LEDdata = 1 << FlashLED;
			FlashTimer = 50;
		} else if (FlashTimer == 25){
			LEDdata = 0;
		}
	}

	if (LEDdata & SW_1)		PORTB |= (1<<LED1); else PORTB &= ~(1<<LED1);
	if (LEDdata & SW_2)		PORTB |= (1<<LED2); else PORTB &= ~(1<<LED2);
	if (LEDdata & SW_3)		PORTB |= (1<<LED3); else PORTB &= ~(1<<LED3);
	if (LEDdata & SW_4)		PORTB |= (1<<LED4); else PORTB &= ~(1<<LED4);

	// A/D変換開始 プリスケーラ 8MHz/64 = 125kHz
	ADCSRA = (1<<ADEN)|(1<<ADSC)|(1<<ADIE)|(1<<ADPS1)|(1<<ADPS2);
}


//*******************************************************
//	A/D変換の割り込みルーチン
//*******************************************************
ISR(ADC_vect) {

	BattVolt = (uint)ADCL;
	BattVolt |= (((uint)ADCH & 3) << 8);
	ADCSRA &= ~(1<<ADEN);	// A/D停止
}


//*******************************************************
//	メインルーチン
//*******************************************************
int main (void) {
	uint	freq;
	uchar	off_count;

	cli();	// 割り込み禁止

	// ポート設定
	PORTB = (1<<PB7)|(1<<PB6)|(1<<SCK)|(1<<MISO);
	DDRB = (1<<LED4)|(1<<LED3)|(1<<LED2)|(1<<LED1);

	PORTD = (1<<SW4)|(1<<SW3)|(1<<SW2)|(1<<SW1)|(1<<SW0)|(1<<TXD)|(1<<RXD);
	DDRD = (1<<POWER);

	PORTC = (1<<PC3)|(1<<PC2)|(1<<PC1);
	DDRC = 0;

#ifdef PRR	// ATmega88/168
	// スリープ動作設定
	SMCR = (1<<SE);			// スリープ許可 アイドルモード

	// タイマ２ CTC動作
	TCCR2A = (1<<WGM21);
	TCCR2B = (1<<CS22)|(1<<CS21)|(1<<CS20);		// CTC動作 プリスケーラ 1/1024
	OCR2A = (CPU_CLOCK/1024/100)-1;		// タイマ 約100Hz
	TIMSK2 = (1<<OCIE2A);	// 比較Ａ割り込み許可

	// 電力カット
	PRR = (1<<PRTIM0)|(1<<PRTIM1)|(1<<PRSPI)|(1<<PRUSART0);

	DIDR0 = (1<<SCL)|(1<<SDA)|(1<<BATT);	// BATTピンデジタル入力禁止

#else	// ATmega8
	// スリープ動作設定
	MCUCR = (1<<SE);			// スリープ許可 アイドルモード

	TCCR2 = (1<<WGM21)|(1<<CS22)|(1<<CS21)|(1<<CS20);	// CTC, 1/1024
	OCR2 = (CPU_CLOCK/1024/100)-1;		// タイマ 約100Hz
	TIMSK = (1<<OCIE2);		// 比較割り込み許可
#endif

	ACSR = (1<<ADC);	// コンパレータ電源OFF

	ADMUX = (1<<REFS1)|(1<<REFS0)|BATT;			// 内部基準電圧 1.1V, A/D ch

	sei();		// 割り込み許可

	// 約0.1秒時間待ち
	wait_timer(10);

	// 電源電圧が低い場合は起動しない
	while(check_batt()){
		wait_timer(1);
	}

	PORTD |= (1<<POWER);	// 電源保持	

	// JP2-1 ON で常にモノラル動作
	R1Default = R1_DEFAULT;
	if (!(PINC & (1<<PC1)))	R1Default |= R1_MONO;

	// EEPROMから前回のチャンネル番号取得
	FMch = eeprom_read_byte((uint8_t *)EEPROM_CH);
	if (FMch > CH_N){
		FMch = 0;
	}
	LEDdata = 1 << (FMch + 1);	// LED点灯
	
	wait_timer(50);		// 電源が安定するまで待つ
	
	// TWI リセット
	TWI_start();
	TWI_send(0xFF);
	TWI_start();
	TWI_stop();

	// AR1000 レジスタ初期値
	write_AR1000(R0, R0_DEFAULT | R0_XO_EN);	// 最初は必ず ENABLE=OFF
	write_AR1000(R1, R1Default | R1_HMUTE);	// ミュート
	write_AR1000(R2, R2_DEFAULT);
	write_AR1000(R3, R3_DEFAULT | R3_VOLUME);
	write_AR1000(R4, 0x0400);
	write_AR1000(R5, 0x28AA);
	write_AR1000(R6, 0x4400);
	write_AR1000(R7, 0x1EE7);
	write_AR1000(R8, 0x7141);
	write_AR1000(R9, 0x007D);
	write_AR1000(R10, 0x82CE);
	write_AR1000(R11, 0x4F55);
	write_AR1000(R12, 0x970C);
	write_AR1000(R13, 0xB845);
	write_AR1000(R14, 0xFC2D);
	write_AR1000(R15, 0x8097);
	write_AR1000(R16, 0x04A1);
	write_AR1000(R17, 0xDF6A);

	write_AR1000(R0, R0_DEFAULT | R0_XO_EN | R0_ENABLE);	// ENABLE
	for (;;){
		wait_timer(10);	// 100ms WAIT
		if (read_AR1000(STATUS) & STATUS_STC)	break;	// STC=H 待ち
	}

	write_AR1000(R1, R1Default);	// ミュート解除

	// SW2とSW4を押しながら電源を入れたら周波数設定モード
	if ((SwFlag & (SW_1|SW_2|SW_3|SW_4)) == (SW_2|SW_4)){
		scan_wave();	// 周波数スキャン＆設定
	}

	LEDdata = 1 << (FMch + 1);
	freq = get_freq(FMch);	// EEPROMから周波数取得
	set_freq(freq);			// 周波数設定

	// スイッチを放すまで待つ
	while (SwFlag){
		wait_timer(1);
	}

	off_count = 0;
	for(;;){

		wait_timer(1);	// 10ms SLEEP

		if (SleepFlag && (SleepTimerM >= SLEEP_MIN)){
			power_off();	// 電源OFF
		}
		
		if (SwFlag & SW_POWER){

			// POWER SW
			if (++off_count >= 10){		// 100ms未満は無視
				LEDdata = 0;	// LED消灯
				// ボタンが放されるまで待つ
				while(SwFlag & SW_POWER){
					wait_timer(1);
					if (off_count < 255)	++off_count;
					if (off_count == 200){	// 2秒長押しでスリープモード
						FlashLED = FMch + 1;	// LED点滅
					}
				}
				if (off_count < 200){	// 2秒未満
					power_off();		// 電源OFF
				} else {
					// スリープモード設定
					SleepTimerDS = SleepTimerS = SleepTimerM = 0;
					SleepFlag = 1;
					FlashLED = 0;
					LEDdata = 1 << (FMch + 1);
				}
			}
		} else if (SwFlag){
			wait_timer(1);

			// 1～4 SW
			if (SwFlag & SW_1){
				FMch = 0;
			} else if (SwFlag & SW_2){
				FMch = 1;
			} else if (SwFlag & SW_3){
				FMch = 2;
			} else if (SwFlag & SW_4){
				FMch = 3;
			}

			if (SwFlag == (SW_3 | SW_4)){
				LEDdata = SW_3 | SW_4;
				freq = FREQ_EXT;
			} else {
				LEDdata = 1 << (FMch + 1);			// LED点灯
				freq = get_freq(FMch);
			}
			set_freq(freq);
			off_count = 0;

			// スイッチが放されるまで待つ
			while(SwFlag){
				wait_timer(1);
			}
		} else {
			off_count = 0;
		}

		// バッテリーチェック
		if (check_batt()){
			power_off();		// 電源OFF
		}
	}
}


//*******************************************************
//	電池電圧チェック
//*******************************************************
char check_batt(void){
	long l;

	cli();
	l = BattVolt;
	sei();
	l *= R8_VALUE + R9_VALUE;
	l /= R8_VALUE;
	l *= ADC_REF;
	l /= 1023;
	if (l < LOW_BATT){
		return 1;
	} else {
		return 0;
	}
}


//*******************************************************
//	電源OFF処理
//*******************************************************
void power_off(void){

	if (eeprom_read_byte((uint8_t *)EEPROM_CH) != FMch){
		eeprom_write_byte((uint8_t *)EEPROM_CH, FMch);	// チャンネル番号保存
	}
	cli();
	LEDdata = 0;
	PORTD &= ~(1<<POWER);		// 電源保持を解除
	for (;;){		// 電源が切れる
		WDR;
		SLEEP;
	}
}


//*******************************************************
//	周波数スキャン＆設定
//*******************************************************
void scan_wave(void){
	uint	memory[CH_N];
	uchar	ch;
	uint	freq;
	uchar	sw;
	uchar	t;
	uchar	r;

	// スイッチを放すまで LED1を点滅
	FlashLED = 1;
	while(SwFlag){
		wait_timer(10);
	}
	wait_timer(10);	// 念のため

	// 設定チャンネルボタン入力待ち
	t = 100;		// 10秒
	while(!(sw = SwFlag)){
		wait_timer(10);
		if (--t == 0){
			FlashLED = 0;	// 時間切れ
			return;
		}
	}

	if (sw & SW_POWER){
		// POWER SWの場合は全チャンネルスキャン
		memset(memory, 0xFF,  sizeof(memory));
		freq = 760;
		ch = 0;
	} else {
		// チャンネルSWの場合はそのチャンネルデータを消去
		// 後のチャンネル周波数を詰める
		for (ch = 0; ch < CH_N; ch++){
			memory[ch] = get_freq(ch);
		}
		for (ch = 0; ch < CH_N; ch++){
			sw >>= 1;
			if (sw & 1){
				break;
			}
		}
		for (; ch < (CH_N - 1); ch++){
			memory[ch] = memory[ch + 1];
		}
		if (memory[ch] >= 1080)	memory[ch] = 760;
		memory[ch] += 3;
		freq = memory[ch];
	}
	FlashLED = ch + 1;

	// スイッチを放すまで待つ
	while(SwFlag){
		wait_timer(10);
	}
	wait_timer(10);	// 念のため
	
	// 周波数スキャン開始
	while(ch < CH_N){

		while (!(r = check_freq(freq))){
			if (SwFlag & SW_POWER)	break;	// POWER SWで中断
			freq++;
			if (freq > 1080)	break;
		}

		if (!r)	break;

		memory[ch++] = freq;

		if (SwFlag & SW_POWER)	break;	// POWER SWで中断

		wait_timer(50);

		FlashLED++;			// 次のLED点滅
		freq += 3;			// 300kHzスキップ
		if (freq > 1080){
			break;
		}
	}
	FlashLED = 0;

	if (!SwFlag){
		// 設定周波数書き込み
		for (ch = 0; ch < CH_N; ch++){
			if (eeprom_read_word((uint16_t *)(EEPROM_FREQ_START + (ch * 2)))
				!= memory[ch]){
				eeprom_write_word((uint16_t *)((EEPROM_FREQ_START + (ch * 2))),
								  memory[ch]);
			}
		}
	}
}


//*******************************************************
//	周波数設定
//*******************************************************
void set_freq(uint freq){

	if ((freq < 760)||(freq > 1080))	return;

	freq -= 690;
	write_AR1000(R2, (R2_DEFAULT & ~(R2_TUNE|R2_CHAN)) + freq);
	write_AR1000(R2, (R2_DEFAULT & ~(R2_TUNE|R2_CHAN)) + freq + R2_TUNE);

	for (;;){
		wait_timer(2);
		if (read_AR1000(STATUS) & STATUS_STC)	break;	// STC
	}
}


//*******************************************************
//	チャンネル番号から周波数取得
//*******************************************************
uint get_freq(uchar ch){
	uint	freq;

	if (ch >= CH_N)	return 0xFFFF;

	// EEPROMから周波数値(100kHz単位)読み込み
	freq = eeprom_read_word((uint16_t *)(EEPROM_FREQ_START + (ch * 2)));

	return freq;
}


//*********************************************************
//	所定の周波数に電波があるか調べる
//*********************************************************
uchar check_freq(uint freq){
	uint	rssi;
	uint	if_cnt;
	uint	reg;
	uchar	i;

	set_freq(freq);		// 周波数設定

	wait_timer(8);		// ちょっと待つ

	rssi = 0;
	if_cnt = 0;
	for (i = 0; i < 4; i++){
		wait_timer(2);		// ちょっと待つ
		reg = read_AR1000(RSSI);
		rssi += (uchar)(reg >> 9);
		if_cnt += reg & RSS_IF_CNT;
	}
	rssi += 2;
	rssi /= 4;
	if_cnt += 2;
	if_cnt /= 4;

	if ((rssi >= RSSI_MIN_LEVEL)&&
		(if_cnt >= 223)&&
		(if_cnt <= 277)){
		return 1;
	} else {
		return 0;
	}
}


//*******************************************************
//	AR1000 レジスタ書き込み(TWI)
//*******************************************************
void write_AR1000(uchar adrs, ushort data){

	TWI_start();
	TWI_send(0x20);			// スレーブアドレス + Write
	TWI_send(adrs);
	TWI_send(data >> 8);
	TWI_send(data & 0x00FF);
	TWI_stop();
}


//*******************************************************
//	AR1000 レジスタ読み込み(TWI)
//*******************************************************
ushort read_AR1000(uchar adrs){
	ushort	data;

	TWI_start();
	TWI_send(0x20);			// スレーブアドレス + Write
	TWI_send(adrs);
	TWI_start();
	TWI_send(0x20 + 1);		// スレーブアドレス + Read
	data = TWI_recv(TWI_ACK);
	data <<= 8;
	data |= TWI_recv(TWI_NACK);
	TWI_stop();

	return data;
}


//*******************************************************
//	Timerを用いた時間待ち
//*******************************************************
void wait_timer(uchar t){

	Timer = t;
	while(Timer){
		WDR;
		SLEEP;
	}
}


//*******************************************************
//	TWI Start
//*******************************************************
char TWI_start(void){
	uint	i;

	TWBR = 32;
	TWSR = 0;		// TWI clock	8MHz / (16 + 2 * TWBR) =  100kHz

	TWCR = (1<<TWINT)|(1<<TWSTA)|(1<<TWEN);	// 開始条件送出
	for (i = 0; i < 10000; i++){
		if (TWCR & (1<<TWINT))	break;
	}

	if ((TWSR & TW_NO_INFO) != TW_START){
		return 1;	// ERROR
	} else {
		return 0;
	}
}


//*******************************************************
//	TWI Send
//*******************************************************
char TWI_send(uchar data){
	uint	i;

	TWDR = data;	// データ
	TWCR = (1<<TWINT)|(1<<TWEN);
	for (i = 0; i < 10000; i++){
		if (TWCR & (1<<TWINT))	break;
	}

	if ((TWSR & TW_NO_INFO) != TW_MT_DATA_ACK){
		return 1;	// ERROR
	} else {
		return 0;
	}
}


//*******************************************************
//	TWI Recieve
//*******************************************************
uchar TWI_recv(uchar ack){
	uint	i;
	uchar	info;

	if (ack){
		TWCR = (1<<TWINT)|(1<<TWEN)|(1<<TWEA);
		info = TW_MR_DATA_ACK;
	} else {
		TWCR = (1<<TWINT)|(1<<TWEN);
		info = TW_MR_DATA_NACK;
	}

	for (i = 0; i < 10000; i++){
		if (TWCR & (1<<TWINT))	break;
	}

	if ((TWSR & TW_NO_INFO) != info){
		return 0;		// ERROR
	} else {
		return TWDR;	// データ
	}
}


//*******************************************************
//	TWI Stop
//*******************************************************
void TWI_stop(void){

	TWCR = (1<<TWINT)|(1<<TWSTO)|(1<<TWEN);
}


